/*
 * _beep_brrr_01.xc
 *
 *  Created on: 2. apr. 2021
 *      Author: teig
 */


#define INCLUDES
#ifdef INCLUDES
    #include <xs1.h>
    #include <platform.h> // slice
    #include <timer.h>    // delay_milliseconds(200), XS1_TIMER_HZ etc
    #include <stdint.h>   // uint8_t
    #include <stdio.h>    // printf
    #include <string.h>   // memcpy
    #include <xccompat.h> // REFERENCE_PARAM(my_app_ports_t, my_app_ports) -> my_app_ports_t &my_app_ports
    #include <iso646.h>   // not etc.
    #include <i2c.h>
    #include <i2s.h>
    #include <xassert.h>

    #include "_version.h" // First this..
    #include "_globals.h" // ..then this

    #include "param.h"
    #include "i2c_client_task.h"
    #include "display_ssd1306.h"
    #include "core_graphics_adafruit_gfx.h"
    #include "core_graphics_font5x8.h"
    #include "_texts_and_constants.h"
    #include "button_press.h"
    #include "pwm_softblinker.h"
    #include "maths.h"
    #include "mic_array.h"
    #include "maths_fix16.h" // Needs stdint.h
    #include "F_Pattern_KnockCome.h"
    #include "mics_in_headset_out.h"
    #include "_Beep_BRRR_01.h"

    #define DO_DEBUG_PRINT 1 // 0 has consequences if (USE_XSCOPE == 1)
    #define DO_WARNINGS    1
    #include "_print_macros_xc.h"
#endif

typedef enum display_screen_name_t {
    SCREEN_INIT,
    SCREEN_MIC,
    // SCREEN_MICS_CONFIG,
    SCREEN_HEADSET_GAIN_DB,
    SCREEN_ABOUT, // LAST SEEN
    SCREEN_DARK   // LAST (NOT SEEN)
} display_screen_name_t;

typedef enum {is_on, is_off} display_state_e;

typedef struct {
    display_state_e       state;
    display_screen_name_t display_screen_name;
    display_screen_name_t display_screen_name_when_into_dark;
    char                  display_ts1_chars [SSD1306_TS1_DISPLAY_VISIBLE_CHAR_LEN]; // 84 chars for display needs 85 char buffer (with NUL) when sprintf is use (use SSD1306_TS1_DISPLAY_ALL_CHAR_LEN for full flexibility)
    int                   sprintf_numchars;
    unsigned              screen_timeouts_since_last_button_countdown; // From NUM_TIMEOUTS_BEFORE_SCREEN_DARK to zero for SCREEN_DARK
    unsigned              cnt;
} display_context_t;

typedef struct {
    bool     now;
    unsigned timeouts_after_last_button_cnt;
    unsigned volume_step_factor; // increasing for every repeat taken, so that going to VOLUME_MIN_DB does not feel "endless"
} repeat_t;

typedef enum {was_none, was_button, was_timeout} last_action_e;

typedef struct {
    bool     valid;
    unsigned index; // [0..BUTTONS_NUM_CLIENTS]
} button_t;

typedef struct {
    button_t        button;
    bool            pressed_ever;
    bool            button_action_taken;
    button_action_t button_action [BUTTONS_NUM_CLIENTS];
    repeat_t        repeat;
    last_action_e   last_action; // AMUX=005
    bool            ignore_left_button_release_no_wake_from_dark; // AMUX=006 new. Since I started with LEFT_BUTTON take on released
    mic_result_t    mic_result;
    unsigned        KnockCome_cnt;                 // debug
    unsigned        KnockCome_OverflowCntDiff_Max; // debug
    uint64_t        KnockCome_OverflowCntDiff_Sum; // debug
} buttons_context_t;

// For set_one_percent_ms and set_sofblink_percentages
#define SOFTBLINK_RESTARTED_ONE_PERCENT_MS        1 //  1 ms goes to 100 in 0.1 seconds -> 5 blinks per second
#define SOFTBLINK_RESTARTED_UNIT_MAX_PERCENTAGE 100
#define SOFTBLINK_RESTARTED_UNIT_MIN_PERCENTAGE   0
//
#define SOFTBLINK_DARK_DISPLAY_ONE_PERCENT_MS    30 // 30 ms goes to 100 in 3.0 seconds
#define SOFTBLINK_DARK_DISPLAY_MAX_PERCENTAGE    40
#define SOFTBLINK_DARK_DISPLAY_MIN_PERCENTAGE    10
//
#define SOFTBLINK_LIT_DISPLAY_ONE_PERCENT_MS     10 // 10 ms goes to 100 in 1.0 seconds
#define SOFTBLINK_LIT_DISPLAY_MAX_PERCENTAGE     80
#define SOFTBLINK_LIT_DISPLAY_MIN_PERCENTAGE     10

#define NUM_TIMEOUTS_PER_SECOND     2
#define NUM_TIMEOUTS_BEFORE_REPEAT (1.5 * NUM_TIMEOUTS_PER_SECOND)  // 1,5 seconds
//      BUTTON_ACTION_PRESSED_FOR_LONG_TIMEOUT_MS must be so long that it does not interfere here

// From makefile
#if (DISPLAY_FAST_DARK==1)
    #if (WARNINGS==1)
        #warning Not standard display to dark timeout
    #endif
    #define NUM_TIMEOUTS_BEFORE_SCREEN_DARK (10 * NUM_TIMEOUTS_PER_SECOND)     // 10 seconds, for screen_timeouts_since_last_button
#elif (DISPLAY_FAST_DARK==0)
    #define NUM_TIMEOUTS_BEFORE_SCREEN_DARK ((5*60) * NUM_TIMEOUTS_PER_SECOND) // 5 minutes, for screen_timeouts_since_last_button
#endif

void do_display_params_zero (display_context_t &display_context) {

    Clear_All_Pixels_In_Buffer();

    for (int index_of_char = 0; index_of_char < NUM_ELEMENTS(display_context.display_ts1_chars); index_of_char++) {
        display_context.display_ts1_chars [index_of_char] = ' ';
    }

    setTextColor(WHITE);
    setCursor(0,0);
    setTextSize(1); // SSD1306_TS1_LINE_CHAR_NUM gives 21 chars per line (but crlf takes two, if used)
}

char button_to_char (const button_t button) {
    char char_ret;

    if (button.valid) {
        char_ret = 'A' + button.index;
    } else {
        char_ret = '?';
    }
    return char_ret;
}

// MUST NOT MODIFY ANY STATE VALUES!
// 0104 kills real-time (see scope picture)
bool // i2c_ok
    Display_screen (
        display_context_t                &display_context,
        buttons_context_t                buttons_context,      // No & no  modification
        headset_gain_context_t           headset_gain_context, // No & no  modification
        client  i2c_internal_commands_if if_i2c_internal_commands) {

    bool i2c_ok = true;
    bool do_display_print = false;

    do_display_params_zero (display_context);

    if (display_context.state == is_on) {
        switch (display_context.display_screen_name) {
            case SCREEN_INIT: {
                debug_print("%s\n", "Display_screen");
                setTextSize(2);
                display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                        "Beep-BRRR");
                display_context.cnt++;
                do_display_print = true;
            } break;

            case SCREEN_MIC: {
                debug_print (
                    "CNT: %u  MIN: %d (%d)  MAX: %d (%d)\n",
                        buttons_context.mic_result.cnt,
                        buttons_context.mic_result.min, MIC_ARRAY_WORD_DB(buttons_context.mic_result.min),
                        buttons_context.mic_result.max, MIC_ARRAY_WORD_DB(buttons_context.mic_result.max));

                setTextSize(1);
                display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                    "MIN %d\n%d dB\nMAX %d\n%d dB",
                        buttons_context.mic_result.min, MIC_ARRAY_WORD_DB(buttons_context.mic_result.min),
                        buttons_context.mic_result.max, MIC_ARRAY_WORD_DB(buttons_context.mic_result.max));
                do_display_print = true;
            } break;

            case SCREEN_HEADSET_GAIN_DB: {
                setTextSize(1);

                // See ...Beep_BRRR v0240 display KnockCome measurements.jpg display photos: KC_THROUGHPUT_TAG

                const unsigned KnockCome_OverflowCntDiff_Mean = (unsigned) (buttons_context.KnockCome_OverflowCntDiff_Sum / (uint64_t) buttons_context.KnockCome_cnt);

                display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                    "%s: %d dB\nKCov: %u\nKCmax: %u %u\nX: %u %u",
                        headset_gain_context.gain_off ? "AV" : "UT",
                        headset_gain_context.amplitude_db_table[headset_gain_context.iof_gain],
                        //buttons_context.mic_result.tmr_timeout_cnt,
                        //buttons_context.mic_result.tmr_immediate_cnt,
                        //buttons_context.mic_result.KnockCome_tics,
                        buttons_context.mic_result.KnockCome_OverflowCnt,
                        buttons_context.KnockCome_OverflowCntDiff_Max,
                        KnockCome_OverflowCntDiff_Mean,
                        (unsigned) buttons_context.KnockCome_OverflowCntDiff_Sum,
                        (unsigned) buttons_context.KnockCome_cnt);
                        //buttons_context.mic_result.used_ticks,
                        //buttons_context.mic_result.max_used_ticks);

                do_display_print = true;
            } break;

            case SCREEN_ABOUT: {
                const char char_OE_str[]          = CHAR_OE_STR; // 
                const char char_right_arrow_str[] = CHAR_RIGHT_ARROW_STR;
                setTextSize(1);
                display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                                      "MicArray board XMOS\nXC KODE %s\nV:%s  xT:%s\n%syvind Teig",
                                      __DATE__,
                                      BEEP_BRRR_VERSION_STR,
                                      XTIMECOMPOSER_VERSION_STR,
                                      char_OE_str);
                   //                                            MicArray XMOS XA
                   //                                            XC CODE JUN 18 2020
                   //                                            V:0.1.75  xT:14.4.1
                   //                                            yvind Teig
                do_display_print = true;
            } break;

            default: {} break;
        }
    } else {
        // is_off : no code
    }

    if (do_display_print) {
        display_print (display_context.display_ts1_chars, display_context.sprintf_numchars); // num chars not including NUL
        i2c_ok = writeToDisplay_i2c_all_buffer (if_i2c_internal_commands);
        // debug_print ("%s\n", i2c_ok ? "ok2" : "err2");
    } else {}

    return i2c_ok;
}

void display_context_init (display_context_t &display_context) {

    display_context.display_screen_name                         = SCREEN_INIT;
    display_context.state                                       = is_on;
    display_context.screen_timeouts_since_last_button_countdown = NUM_TIMEOUTS_BEFORE_SCREEN_DARK;
    display_context.cnt                                         = 0;
}

void buttons_context_init (buttons_context_t &buttons_context) {
    buttons_context.button.valid = false;
    //
    buttons_context.KnockCome_cnt                 = 1; // div by zero not ok
    buttons_context.KnockCome_OverflowCntDiff_Max = 0;
    buttons_context.KnockCome_OverflowCntDiff_Sum = 0;
}

// [[combinable]]
void Handle_Beep_BRRRR_task_b (
         server button_if_1              if_button [BUTTONS_NUM_CLIENTS],
         out port                        leds_00_07,
         client softblinker_if           if_softblinker,
         client i2c_internal_commands_if if_i2c_internal_commands,
         client i2c_general_commands_if  if_i2c_general_commands,
         out port                        p_display_notReset,
         out port                        p_scope_orange,
         chanend                         ch_ab_bidir, // ch_ab_bidir_t
         streaming chanend               ch_ba_knock) // ch_ba_knock_t
{
    #if (USE_XSCOPE == 1)
        delay_milliseconds (DELAY_FIRST_PRINT_MS);
    #endif
    debug_print("%s\n", "Handle_Beep_BRRRR_task_b started");

    unsigned one_percent_ms = SOFTBLINK_DARK_DISPLAY_ONE_PERCENT_MS;
    unsigned cnt = 0;
    timer    tmr;
    time32_t time_ticks;

    display_context_t      display_context;
    buttons_context_t      buttons_context;
    ch_ab_bidir_t          data_ch_ab_bidir;
    headset_gain_context_t headset_gain_context;
    ch_ba_knock_t          data_ch_ba_knock;

    menu_result_t menu_result;

    set_headset_gain (do_headset_gain_default, headset_gain_context);

    menu_result.headset_gain_updated = false; // So other end of ch_ab_bidir would not use any value
    menu_result.directive_mics_valid = false; // So I will nit use any value

    // STARTUP
    if_softblinker.set_one_percent_ms (one_percent_ms);
    if_softblinker.set_sofblink_percentages (SOFTBLINK_RESTARTED_UNIT_MAX_PERCENTAGE, SOFTBLINK_RESTARTED_UNIT_MIN_PERCENTAGE);

    buttons_context_init (buttons_context);
    display_context_init (display_context);

    Adafruit_GFX_constructor   (SSD1306_LCDWIDTH, SSD1306_LCDHEIGHT);
    Adafruit_SSD1306_i2c_begin (if_i2c_internal_commands, p_display_notReset);
    Display_screen             (display_context, buttons_context, headset_gain_context, if_i2c_internal_commands);

    DEBUG_1PORT_LOW (p_scope_orange);

    tmr :> time_ticks; // Immeddiately;

    while(1) {
        [[ordered]] // Ok, but not strictly necessary
        select {
            case ch_ba_knock :> data_ch_ba_knock : {
                DEBUG_1PORT_HIGH (p_scope_orange);

                // NOT NEEDED since this event is always allowed and the next knock-come sequence is atomic:
                // KnockCome_State_e KnockCome_State;

                xassert (data_ch_ba_knock.KnockCome_Message_Type == KC_TYP_SM_KNOCK);

                #if (DEBUG_F_PATTERN_KNOCKCOME == 1)
                    debug_print ("%s\n", "KC-M: Knock received");
                #endif

                // Build response
                data_ch_ab_bidir.source = task_b;
                // No need to add any data here, so       KC_TYP_MS_COME_DATA is never used:
                data_ch_ab_bidir.KnockCome_Message_Type = KC_TYP_MS_COME;

                // =============================================================
                // INSIDE THIS CASE CONTIUE WITH THIS ATOMIC KNOCK-COME SEQUENCE
                // =============================================================


                ch_ab_bidir <: data_ch_ab_bidir; // SEND

                #if (DEBUG_F_PATTERN_KNOCKCOME == 1)
                    debug_print ("%s\n", "KC-M: Come sent");
                #endif

                DEBUG_1PORT_LOW (p_scope_orange);

                ch_ab_bidir :> data_ch_ab_bidir; // RECEIVE

                DEBUG_1PORT_HIGH (p_scope_orange);

                #if (DEBUG_F_PATTERN_KNOCKCOME == 1)
                    debug_print ("%s\n", "KC-M: Data received. Finished");
                #endif
                // =============================================================

                // Analyze reponse
                xassert (data_ch_ab_bidir.source == task_a);
                xassert (data_ch_ba_knock.KnockCome_Message_Type == KC_TYP_SM_KNOCK);

                buttons_context.mic_result = data_ch_ab_bidir.data.mic_result;

                if (buttons_context.mic_result.KnockCome_OverflowCnt > 0) {
                    if ((buttons_context.mic_result.KnockCome_OverflowCnt % 100) == 0) {
                        debug_print ("OC %u\n", buttons_context.mic_result.KnockCome_OverflowCnt);
                    } else {}
                } else {}

                buttons_context.KnockCome_cnt++;
                buttons_context.KnockCome_OverflowCntDiff_Max =
                        max (
                                buttons_context.mic_result.KnockCome_OverflowCntDiff,
                                buttons_context.KnockCome_OverflowCntDiff_Max);

                buttons_context.KnockCome_OverflowCntDiff_Sum +=  (uint64_t) buttons_context.mic_result.KnockCome_OverflowCntDiff;

                // DISPLAY HANDLING
                //
                // TIMING of mics_input_task not disturbed because the channel input "returns" immediately
                /*
                display_context.state = is_on;
                display_context.screen_timeouts_since_last_button_countdown = NUM_TIMEOUTS_BEFORE_SCREEN_DARK;
                display_context.display_screen_name = SCREEN_MIC;
                Display_screen (display_context, buttons_context, headset_gain_context, if_i2c_internal_commands);
                */

                DEBUG_1PORT_LOW (p_scope_orange);
            } break;

            case tmr when timerafter (time_ticks) :> void : {
                time_ticks += (XS1_TIMER_HZ/NUM_TIMEOUTS_PER_SECOND);
                buttons_context.button.valid = false;

                if (display_context.screen_timeouts_since_last_button_countdown > 0) {
                    display_context.screen_timeouts_since_last_button_countdown--;
                } else { // == 0
                    display_context.state = is_off;
                    Display_screen (display_context, buttons_context, headset_gain_context, if_i2c_internal_commands);
                }

            } break;

            // If the button is "stable", react when the I/O pin changes value
            case if_button[unsigned iof_button].button (const button_action_t button_action) : {
                //
                // 0 BUTTON_ACTION_VOID,
                // 1 BUTTON_ACTION_PRESSED,
                // 2 BUTTON_ACTION_PRESSED_FOR_LONG, // BUTTON_ACTION_PRESSED_FOR_LONG_TIMEOUT_MS
                // 3 BUTTON_ACTION_RELEASED          // Not after BUTTON_ACTION_PRESSED_FOR_LONG

                one_percent_ms = (SOFTBLINK_RESTARTED_ONE_PERCENT_MS * (iof_button+1));

                #if (USE_XSCOPE == 1)
                    // Would have destroyd all timing
                #elif (USE_XSCOPE == 0)
                    debug_print("A Handle_Beep_BRRRR_task_b [%u]=%u\n", iof_button, button_action);
                #endif

                if_softblinker.set_one_percent_ms (one_percent_ms);

                cnt++;

                if (button_action == BUTTON_ACTION_PRESSED) {

                    display_context.state = is_on;
                    display_context.screen_timeouts_since_last_button_countdown = NUM_TIMEOUTS_BEFORE_SCREEN_DARK;
                    buttons_context.button.valid = true;
                    buttons_context.button.index = iof_button;
                    //
                    switch (iof_button) {
                        case IOF_BUTTON_LEFT: {
                            set_headset_gain (do_headset_gain_zero, headset_gain_context);
                        } break;
                        case IOF_BUTTON_CENTER: {
                            set_headset_gain (do_headset_gain_decrease, headset_gain_context);
                        } break;
                        case IOF_BUTTON_RIGHT: {
                            set_headset_gain (do_headset_gain_increase, headset_gain_context);
                        } break;
                        case IOF_BUTTON_FAR_RIGHT: {
                            set_headset_gain (do_headset_gain_default, headset_gain_context);
                        } break;
                        // default: fault by xC runtime
                    }
                    display_context.display_screen_name = SCREEN_HEADSET_GAIN_DB;
                    //
                    menu_result.clear_debug_vars = (iof_button == IOF_BUTTON_LEFT); // Return value too
                    //
                    if (menu_result.clear_debug_vars) {
                        buttons_context_init (buttons_context);
                    } else {}

                    menu_result.headset_gain         = headset_gain_context.gain;
                    menu_result.headset_gain_off     = headset_gain_context.gain_off;
                    menu_result.headset_gain_updated = true;

                    // KC_THROUGHPUT_TAG this is the one that seems to have a 100 ms delay
                    Display_screen (display_context, buttons_context, headset_gain_context, if_i2c_internal_commands);

                    data_ch_ab_bidir.KnockCome_Message_Type = KC_TYP_MS_NONE_DATA;
                    data_ch_ab_bidir.data.menu_result = menu_result;
                    data_ch_ab_bidir.source = task_b;

                    ch_ab_bidir <: data_ch_ab_bidir;

                    menu_result.headset_gain_updated = false; // Not really needed
                }
            } break;
        }
    }
};
